__all__ = [
    "RPCException",
]


class RPCException(Exception):
    """Base class for all RPC exceptions."""

    def __init__(self, code: int, message: str):
        super().__init__(message)
        self.code = code

    def __str__(self):
        return f"RPC request error with code {self.code} [{self.args[0]}]"



import json
import os
import uuid
from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from typing import Dict
from typing import List
from typing import Literal
from typing import Tuple
from typing import Union

import requests
from web3 import Web3
from web3.types import Wei

from decodex.exceptions import RPCException
from decodex.type import AccountBalanceChanged
from decodex.type import Log
from decodex.type import RawTraceCallResponse
from decodex.type import RawTraceCallResult
from decodex.type import Tx


class BaseSearcher(ABC):
    @abstractmethod
    def get_tx(self, txhash: str) -> Tx:
        """
        Search a transaction by its hash. Return a TxDict.
        """
        raise NotImplementedError

    @abstractmethod
    def simluate_tx(
        self,
        from_address: str,
        to_address: str,
        value: Wei,
        data: str,
        block: Union[int, Literal["latest"]] = "latest",
        gas: Union[Wei, Literal["auto"]] = "auto",
        gas_price: Union[Wei, Literal["auto"]] = "auto",
        timeout: int = 120,
    ) -> Tx:
        """
        Simluates a transaction and returns the trace result.

        Params
        ------
        from_address: str
            The address of the sender.
        to_address: str
            The address of the receiver.
        value: Wei
            The value of the transaction. In wei (int).
        data: str
            The data of the transaction. 0x prefixed.
        gas: Union[Wei, Literal["auto"]]
            The gas limit of the transaction. In wei (int). If "auto", the gas limit will be estimated.
        gas_price: Union[Wei, Literal["auto"]]
            The gas price of the transaction. In wei (int). If "auto", the gas price will be estimated.
        timeout: str
            The timeout of the transaction. In seconds (int).
        """
        raise NotImplementedError


class Web3Searcher(BaseSearcher):
    def __init__(self, provider: str = None) -> None:
        """
        Initialize a Web3Searcher with a Web3 instance or a Web3 http provider URI.

        Params
        ------
        provider: str
            The Web3 http provider URI. If None, the environment variable WEB3_PROVIDER_URI will be used. Otherwise, the default value is "http://localhost:8545".
        """
        if provider is None:
            provider = os.getenv("WEB3_PROVIDER_URI", "http://localhost:8545")

        if isinstance(provider, str):
            w3 = Web3(Web3.HTTPProvider(provider))
            self.provider = provider
            self.web3 = w3
        else:
            raise TypeError("provider must be a Web3 http provider URI")

    def get_tx(self, txhash: str, *, max_workers: int = 2, show_revert_reason: bool = True) -> Tx:
        assert max_workers > 0, "max_workers must be positive"
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            tx = executor.submit(self.web3.eth.get_transaction, txhash)
            tx_receipt = executor.submit(self.web3.eth.get_transaction_receipt, txhash)
            tx, tx_receipt = tx.result(), tx_receipt.result()

        blk = self.web3.eth.get_block(tx_receipt["blockNumber"])
        logs: list[Log] = [
            {
                "address": log["address"],
                "topics": [t.hex() for t in log["topics"]],
                "data": log["data"].hex(),
            }
            for log in tx_receipt["logs"]
        ]

        status = tx_receipt["status"]
        reason = ""
        if status == 0 and show_revert_reason:
            try:
                self.web3.eth.call(
                    {
                        "from": tx["from"],
                        "to": tx["to"],
                        "value": tx["value"],
                        "data": tx["input"],
                    },
                    tx_receipt["blockNumber"],
                )
            except Exception as e:
                reason = str(e)

        to_addr = tx["to"]
        from_addr = tx["from"]
        gas_used = tx_receipt["gasUsed"]
        gas_price = tx["gasPrice"]
        value = tx["value"]

        eth_balance_changes = {from_addr: {"ETH": -value, "Gas Fee": -gas_used * gas_price}}

        # If to_address is an eoa, add the value to the balance change
        if to_addr and status == 1 and self.web3.eth.get_code(to_addr).hex() == "0x":
            if to_addr not in eth_balance_changes:
                eth_balance_changes[to_addr] = {}
            eth_balance_changes[to_addr]["ETH"] = value

        return {
            "txhash": tx_receipt["transactionHash"].hex(),
            "from": from_addr,
            "to": tx["to"],
            "contract_created": tx_receipt["contractAddress"],
            "block_number": tx_receipt["blockNumber"],
            "block_timestamp": blk["timestamp"],
            "value": value,
            "gas_used": gas_used,
            "gas_price": gas_price,
            "input": tx["input"].hex(),
            "status": status,
            "reason": reason,
            "logs": logs,
            "eth_balance_changes": eth_balance_changes,
        }

    def _parse_calls(self, result: RawTraceCallResult) -> Tuple[List[Log], Dict[str, int]]:
        """
        Recursively parse the calls and return logs and balance changes.
        Balance changes are represented as a dict of address and value in Wei.
        """
        logs: List[Log] = []
        balance_changes = defaultdict(int)
        for call in result.get("calls", []):
            if "logs" in call:
                logs += call["logs"]
            if "value" in call:
                value = int(call["value"], 16)
                if value != 0:
                    balance_changes[call["from"]] -= value
                    balance_changes[call["to"]] += value
            if "calls" in call:
                _logs, _balance_changes = self._parse_calls(call)
                logs += _logs
                for addr, value in _balance_changes.items():
                    balance_changes[addr] += value
        return logs, balance_changes

    def simluate_tx(
        self,
        from_address: str,
        to_address: str,
        value: Wei,
        data: str,
        block: Union[int, Literal["latest"]] = "latest",
        gas: Union[Wei, Literal["auto"]] = "auto",
        gas_price: Union[Wei, Literal["auto"]] = "auto",
        timeout: int = 120,
    ) -> Tx:
        """
        Simluates a transaction and returns the trace result.
        """
        assert isinstance(block, int) or block == "latest", "block must be an integer or 'latest'"
        assert isinstance(gas, int) or gas == "auto", "gas must be an integer or 'auto'"
        assert isinstance(gas_price, int) or gas_price == "auto", "gas_price must be an integer or 'auto'"
        assert timeout > 0, "timeout must be positive"

        blk = self.web3.eth.get_block(block)
        if gas == "auto":
            gas = self.web3.eth.estimate_gas(
                {
                    "from": from_address,
                    "to": to_address,
                    "value": value,
                    "data": data,
                }
            )
            gas = int(gas * 1.5)
        if gas_price == "auto":
            gas_price = self.web3.eth.gas_price
        params = [
            {
                "from": from_address,
                "to": to_address,
                "value": hex(value),
                "data": data,
                "gas": hex(gas),
                "gasPrice": hex(gas_price),
            },
            "latest",
            {
                "timeout": f"{timeout}s",
                "tracer": "callTracer",
                "tracerConfig": {"withLog": True},
            },
        ]
        uid = str(uuid.uuid4())
        payload = {"jsonrpc": "2.0", "method": "debug_traceCall", "params": params, "id": uid}
        headers = {"Content-Type": "application/json"}

        resp: RawTraceCallResponse = requests.request(
            method="POST", url=self.provider, headers=headers, data=json.dumps(payload)
        ).json()

        if resp.get("error", {}):
            raise RPCException(resp["error"]["code"], resp["error"]["message"])

        logs, account_balance = self._parse_calls(resp["result"])
        if len(account_balance) == 0:
            eth_balance_changes = {from_address: {"ETH": -value, "Gas Fee": 0}}
        else:
            eth_balance_changes = {addr: {"ETH": wei, "Gas Fee": 0} for addr, wei in account_balance.items()}

        if to_address and self.web3.eth.get_code(to_address).hex() == "0x":
            if to_address not in eth_balance_changes:
                eth_balance_changes[to_address] = {}
            eth_balance_changes[to_address]["ETH"] = value

        result = resp.get("result", {})
        tx: Tx = {
            "txhash": "Simulation result",
            "from": from_address,
            "to": to_address,
            "contract_created": None,
            "block_number": blk["number"],
            "block_timestamp": blk["timestamp"],
            "value": value,
            "gas_used": int(result["gasUsed"], 16),
            "gas_price": gas_price,
            "input": data,
            "status": 1,
            "reason": "",
            "logs": logs,
            "eth_balance_changes": eth_balance_changes,
        }
        return tx


class SearcherFactory:
    @staticmethod
    def create(
        searcher_type: Literal["web3"],
        uri: str = None,
    ) -> BaseSearcher:
        if searcher_type == "web3":
            return Web3Searcher(provider=uri)
        else:
            raise NotImplementedError



from .dex import BancorEV3Events
from .dex import CurveV2Events
from .dex import Events
from .dex import UniswapV2Events
from .dex import UniswapV3Events
from .lending import AAVEV2Events
from .lending import AAVEV3Events
from .lending import CompoundV3Events
from .tokens import ERC20Events
from .tokens import ERC721Events

__all__ = [
    "Events",
    "ERC20Events",
    "ERC721Events",
    "UniswapV2Events",
    "UniswapV3Events",
    "BancorEV3Events",
    "CurveV2Events",
    "AAVEV2Events",
    "AAVEV3Events",
    "CompoundV3Events",
]



import os
import pathlib
import tempfile
import warnings
from typing import Dict

import requests
from tqdm import tqdm

from .callbacks import ChecksumPostCallback
from .callbacks import GithubLFSBeforeCallback
from .callbacks import GithubRawBeforeCallback
from .callbacks import GzipPostCallback

warnings.filterwarnings("ignore")


def _get_github_url(save_path: str, org: str, repo: str, branch: str, path: str, is_lfs: bool) -> str:
    if is_lfs:
        cls = GithubLFSBeforeCallback
    else:
        cls = GithubRawBeforeCallback

    return cls.before_download(
        save_path=save_path,
        org=org,
        repo=repo,
        branch=branch,
        path=path,
    )


def download_from_url(url: str, save_path: str, verify_ssl: bool, retry_with_proxy: bool = True) -> None:
    try:
        response = requests.get(url=url, verify=verify_ssl, stream=True)
        file_size = int(response.headers.get("content-length", 0))
        with tqdm(total=file_size, unit="iB", unit_scale=True) as pbar:
            with open(save_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)
                        pbar.update(len(chunk))
    except Exception as e:
        if not retry_with_proxy:
            raise e
        proxy = os.getenv("PROXY_URL", None)
        if proxy:
            download_from_url(url, save_path, verify_ssl, retry_with_proxy=False)


def download_github_file(
    save_path: str,
    org: str,
    repo: str,
    branch: str,
    path: str,
    is_lfs: bool = False,
    verify_ssl: bool = False,
    use_tempfile: bool = False,
) -> None:
    # Create parent directory if not exist
    pathlib.Path(save_path).parent.mkdir(parents=True, exist_ok=True)

    try:
        url = _get_github_url(save_path, org, repo, branch, path, is_lfs)
    except FileExistsError as e:
        print(f"Skip Downloading: {e}")
        return
    except Exception as e:
        print(f"Error getting download URL: {e}")
        return

    try:
        src_path = tempfile.mktemp() if use_tempfile else save_path
        download_from_url(url, src_path, verify_ssl)
    except Exception as e:
        print(f"Error downloading file: {e}")
        return

    post_cls = GzipPostCallback if is_lfs else ChecksumPostCallback
    try:
        post_cls.post_download(src_path, save_path)
    except Exception as e:
        print(f"Error in post-download steps: {e}")
        return



from typing import Optional
from typing import Tuple

from multicall import Multicall

from .dex import Events
from decodex.convert.address import AddrTagger
from decodex.type import Action
from decodex.type import BorrowAction
from decodex.type import DepositAction
from decodex.type import DisableCollateralAction
from decodex.type import EnableCollateralAction
from decodex.type import EventHandleFunc
from decodex.type import EventPayload
from decodex.type import FlashloanAction
from decodex.type import RepayAction
from decodex.type import SupplyAction
from decodex.type import WithdrawAction


class AAVEV2Events(Events):
    def __init__(
        self,
        mc: Multicall,
        tagger: AddrTagger,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(mc, tagger, *args, **kwargs)

    def deposit(self) -> Tuple[str, EventHandleFunc]:
        # Deposit (index_topic_1 address reserve, address user, index_topic_2 address onBehalfOf, uint256 amount, index_topic_3 uint16 referral)
        text_sign = "Deposit(address,address,address,uint256,uint16)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["params"]["reserve"]
            token = self._erc20_svc.get_erc20(token_addr)
            if token is None:
                return None
            amount = payload["params"]["amount"] / 10 ** token["decimals"]
            pool, token = self._tagger([payload["address"], token])
            return DepositAction(
                pool=pool,
                token=token,
                amount=amount,
            )

        return text_sign, decoder

    def borrow(self) -> Tuple[str, EventHandleFunc]:
        # Borrow (index_topic_1 address reserve, address user, index_topic_2 address onBehalfOf, uint256 amount, uint256 borrowRateMode, uint256 borrowRate, index_topic_3 uint16 referral)
        text_sign = "Borrow(address,address,address,uint256,uint256,uint256,uint16)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["params"]["reserve"]
            token = self._erc20_svc.get_erc20(token_addr)
            if token is None:
                return None
            amount = payload["params"]["amount"] / 10 ** token["decimals"]
            pool, token = self._tagger([payload["address"], token])
            return BorrowAction(
                pool=pool,
                token=token,
                amount=amount,
            )

        return text_sign, decoder

    def withdraw(self) -> Tuple[str, EventHandleFunc]:
        # Withdraw (index_topic_1 address reserve, index_topic_2 address user, index_topic_3 address to, uint256 amount)
        text_sign = "Withdraw(address,address,address,uint256)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["params"]["reserve"]
            token = self._erc20_svc.get_erc20(token_addr)
            if token is None:
                return None
            amount = payload["params"]["amount"] / 10 ** token["decimals"]
            pool, token = self._tagger([payload["address"], token])
            return WithdrawAction(
                pool=pool,
                token=token,
                amount=amount,
            )

        return text_sign, decoder

    def repay(self) -> Tuple[str, EventHandleFunc]:
        # Repay (index_topic_1 address reserve, index_topic_2 address user, index_topic_3 address repayer, uint256 amount)
        text_sign = "Repay(address,address,address,uint256)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["params"]["reserve"]
            token = self._erc20_svc.get_erc20(token_addr)
            if token is None:
                return None
            amount = payload["params"]["amount"] / 10 ** token["decimals"]
            pool, token = self._tagger([payload["address"], token])
            return RepayAction(
                pool=pool,
                token=token,
                amount=amount,
            )

        return text_sign, decoder

    def flashloan(self) -> Tuple[str, EventHandleFunc]:
        # FlashLoan (index_topic_1 address target, index_topic_2 address initiator, index_topic_3 address asset, uint256 amount, uint256 premium, uint16 referralCode)
        text_sign = "FlashLoan(address,address,address,uint256,uint256,uint16)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["params"]["asset"]
            token = self._erc20_svc.get_erc20(token_addr)
            amount = payload["params"]["amount"] / 10 ** token["decimals"]
            pool, token = self._tagger([payload["address"], token])
            return FlashloanAction(
                pool=pool,
                token=token,
                amount=amount,
            )

        return text_sign, decoder


class AAVEV3Events(Events):
    def __init__(
        self,
        mc: Multicall,
        tagger: AddrTagger,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(mc, tagger, *args, **kwargs)

    def supply(self) -> Tuple[str, EventHandleFunc]:
        # Supply (index_topic_1 address reserve, address user, index_topic_2 address onBehalfOf, uint256 amount, index_topic_3 uint16 referralCode)
        text_sig = "Supply(address,address,address,uint256,uint16)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["params"]["reserve"]
            token = self._erc20_svc.get_erc20(token_addr)
            amount = payload["params"]["amount"] / 10 ** token["decimals"]
            pool, token = self._tagger([payload["address"], token])
            return SupplyAction(
                pool=pool,
                token=token,
                amount=amount,
            )

        return text_sig, decoder

    def borrow(self) -> Tuple[str, EventHandleFunc]:
        # Borrow (index_topic_1 address reserve, address user, index_topic_2 address onBehalfOf, uint256 amount, uint8 interestRateMode, uint256 borrowRate, index_topic_3 uint16 referralCode)
        text_sig = "Borrow(address,address,address,uint256,uint8,uint256,uint16)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["params"]["reserve"]
            # (token_decimals,) = self._get_token_decimals(token_addr)
            token = self._erc20_svc.get_erc20(token_addr)
            if token is None:
                return None
            amount = payload["params"]["amount"] / 10 ** token["decimals"]
            pool, token = self._tagger([payload["address"], token])
            return BorrowAction(
                pool=pool,
                token=token,
                amount=amount,
            )

        return text_sig, decoder

    # def withdraw(self) -> Tuple[str, EventHandleFunc]:
    # pass (Duplicate with AAVEV2Events)

    def flashloan(self) -> Tuple[str, EventHandleFunc]:
        # FlashLoan(address indexed target, address initiator, address indexed asset, uint256 amount, DataTypes.InterestRateMode interestRateMode, uint256 premium, uint16 indexed referralCode);
        text_sig = "FlashLoan(address,address,address,uint256,uint8,uint256,uint16)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["params"]["asset"]
            token = self._erc20_svc.get_erc20(token_addr)
            amount = payload["params"]["amount"] / 10 ** token["decimals"]
            pool, token = self._tagger([payload["address"], token])
            return FlashloanAction(
                pool=pool,
                token=token,
                amount=amount,
            )

        return text_sig, decoder

    def repay(self) -> Tuple[str, EventHandleFunc]:
        # event Repay(address indexed reserve, address indexed user, address indexed repayer, uint256 amount, bool useATokens);
        text_sig = "Repay(address,address,address,uint256,bool)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["params"]["reserve"]
            token = self._erc20_svc.get_erc20(token_addr)
            amount = payload["params"]["amount"] / 10 ** token["decimals"]
            pool, token = self._tagger([payload["address"], token])
            return RepayAction(
                pool=pool,
                token=token,
                amount=amount,
            )

        return text_sig, decoder

    def reserve_used_as_collateral_enabled(self) -> Tuple[str, EventHandleFunc]:
        # ReserveUsedAsCollateralEnabled (index_topic_1 address reserve, index_topic_2 address user)
        text_sig = "ReserveUsedAsCollateralEnabled(address,address)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["params"]["reserve"]
            token = self._erc20_svc.get_erc20(token_addr)
            if token is None:
                return None
            pool, token = self._tagger([payload["address"], token])
            return EnableCollateralAction(
                pool=pool,
                token=token,
            )

        return text_sig, decoder

    def reserve_used_as_collateral_disabled(self) -> Tuple[str, EventHandleFunc]:
        # ReserveUsedAsCollateralDisabled (index_topic_1 address reserve, index_topic_2 address user)
        text_sig = "ReserveUsedAsCollateralDisabled(address,address)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["params"]["reserve"]
            token = self._erc20_svc.get_erc20(token_addr)
            if token is None:
                return None
            pool, token = self._tagger([payload["address"], token])
            return DisableCollateralAction(
                pool=pool,
                token=token,
            )

        return text_sig, decoder


class CompoundV3Events(Events):
    def __init__(
        self,
        mc: Multicall,
        tagger: AddrTagger,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(mc, tagger, *args, **kwargs)

    def supply_collateral(self) -> Tuple[str, EventHandleFunc]:
        # SupplyCollateral (index_topic_1 address from, index_topic_2 address dst, index_topic_3 address asset, uint256 amount)
        text_sig = "SupplyCollateral(address,address,address,uint256)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["params"]["asset"]
            token = self._erc20_svc.get_erc20(token_addr)
            if token is None:
                return None
            amount = payload["params"]["amount"] / 10 ** token["decimals"]
            dst = payload["params"]["dst"]
            return SupplyAction(
                pool=dst,
                token=token,
                amount=amount,
            )

        return text_sig, decoder

    def withdraw(self) -> Tuple[str, EventHandleFunc]:
        # Withdraw (index_topic_1 address src, index_topic_2 address to, uint256 amount)
        text_sig = "Withdraw(address,address,uint256)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["address"]
            token = self._erc20_svc.get_erc20(token_addr)
            if token is None:
                return None
            params = payload["params"]
            amount = params["__idx_2"] / 10 ** token["decimals"]
            src = params["__idx_0"]
            dst = params["__idx_1"]
            return WithdrawAction(
                pool=src,
                token=token,
                amount=amount,
                receiver=dst,
            )

        return text_sig, decoder

    def supply(self) -> Tuple[str, EventHandleFunc]:
        # Supply (index_topic_1 address from, index_topic_2 address dst, uint256 amount)
        text_sig = "Supply(address,address,uint256)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["address"]
            token = self._erc20_svc.get_erc20(token_addr)
            if token is None:
                return None
            params = payload["params"]
            amount = params["__idx_2"] / 10 ** token["decimals"]
            dst = params["__idx_1"]
            return SupplyAction(
                pool=dst,
                token=token,
                amount=amount,
            )

        return text_sig, decoder



from threading import Lock
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

import diskcache
from cachetools import cached
from cachetools import LRUCache
from eth_utils.address import to_checksum_address
from multicall import Call
from multicall import Multicall
from web3 import Web3
from web3.exceptions import BadFunctionCallOutput
from web3.exceptions import ContractLogicError

from decodex.constant import DECODEX_DIR
from decodex.constant import NULL_ADDRESS_0x0
from decodex.constant import NULL_ADDRESS_0xF
from decodex.type import ERC20Compatible


class ERC20TokenService:
    def __init__(self, mc: Multicall, *, cache_path: Optional[str] = None) -> None:
        self._mc = mc
        self._cache = diskcache.Cache(cache_path or DECODEX_DIR.joinpath("erc20"))

    @cached(cache=LRUCache(maxsize=131072), lock=Lock())
    def get_erc20(
        self,
        address: str,
        block_number: Union[int, Literal["latest"]] = "latest",
        *,
        strict: bool = True,
    ) -> Optional[ERC20Compatible]:
        """
        Get ERC20 token information from address.

        Parameters
        ----------
        address : str
            Token address.
        block_number : int or "latest", optional
            Block number to query, by default "latest".
        strict : bool, optional
            if True, return None when token is not found or not ERC20 compatible, by default True.
        """
        token = self._cache.get(address)
        if token is not None:
            return token

        if address in {NULL_ADDRESS_0x0, NULL_ADDRESS_0xF}:
            return {
                "name": "Platform Token",
                "address": address,
                "contract_name": None,
                "symbol": "ETH",
                "decimals": 18,
                "labels": [],
            }

        response: Dict[str, Any] = self._mc.agg(
            [
                Call(
                    target=address,
                    function="name()(string)",
                    request_id=f"{address}-name",
                ),
                Call(
                    target=address,
                    function="symbol()(string)",
                    request_id=f"{address}-symbol",
                ),
                Call(
                    target=address,
                    function="decimals()(uint8)",
                    request_id=f"{address}-decimals",
                ),
            ],
            block_id=block_number,
            as_dict=True,
        )

        name = response.get(f"{address}-name", None)
        symbol = response.get(f"{address}-symbol", None)
        decimals = response.get(f"{address}-decimals", None)

        if strict and (name is None or symbol is None or decimals is None):
            return None

        rtn = {
            "name": None,
            "address": address.lower(),
            "contract_name": name,
            "decimals": decimals,
            "symbol": symbol,
            "labels": [],
        }

        self._cache.set(address, rtn)

        return rtn

    def batch_get_erc20(
        self,
        addresses: List[str],
        block_number: Union[int, Literal["latest"]] = "latest",
    ) -> List[Optional[ERC20Compatible]]:
        return [self.get_erc20(address, block_number=block_number) for address in addresses]



from typing import Optional
from typing import Tuple

from multicall import Multicall

from .dex import Events
from decodex.convert.address import AddrTagger
from decodex.type import Action
from decodex.type import EventHandleFunc
from decodex.type import EventPayload
from decodex.type import TransferAction


class ERC20Events(Events):
    """
    Standard ERC20 token events.
    """

    def __init__(
        self,
        mc: Multicall,
        tagger: AddrTagger,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(mc, tagger, *args, **kwargs)

    def transfer(self) -> Tuple[str, EventHandleFunc]:
        # Transfer (index_topic_1 address from, index_topic_2 address to, uint256 value)
        text_sig = "Transfer(address,address,uint256)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token_addr = payload["address"]
            token = self._erc20_svc.get_erc20(token_addr)
            params = payload["params"]
            amount = params["value"] / 10 ** token["decimals"]
            token, sender, receiver = self._tagger([token, params["from"], params["to"]])
            return TransferAction(
                sender=sender,
                receiver=receiver,
                token=token,
                amount=amount,
            )

        return text_sig, decoder


class ERC721Events(Events):
    def __init__(
        self,
        mc: Multicall,
        tagger: AddrTagger,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(mc, tagger, *args, **kwargs)



import os
from textwrap import indent
from typing import Literal
from typing import Union

import click
import pyfiglet
from colorama import Fore
from colorama import Style
from jinja2 import Template
from tabulate import tabulate

from decodex.constant import DECODEX_DIR
from decodex.installer import download_github_file
from decodex.translate import Translator
from decodex.utils import fmt_addr
from decodex.utils import fmt_blktime
from decodex.utils import fmt_gas
from decodex.utils import fmt_status
from decodex.utils import fmt_value

CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"],
)

WELCOME_MESSAGE = f"{Fore.YELLOW}{pyfiglet.figlet_format('decodex')}{Style.RESET_ALL}"

VERSION_INFO = f"version  |  %(version)s \n"


@click.group(
    context_settings=CONTEXT_SETTINGS,
    invoke_without_command=True,
    no_args_is_help=True,
)
@click.version_option(
    package_name="decodex",
    message=f"{Fore.YELLOW}{pyfiglet.figlet_format('welcome to decodex')}{Style.RESET_ALL} {VERSION_INFO}",
)
def cli():
    pass


@cli.command(help="download tags and signatures for a chain")
@click.argument("chain", default="ethereum", type=click.Choice(["ethereum"]))
@click.option("--verify-ssl", is_flag=True, help="Verify SSL", default=False)
def download(chain: str, verify_ssl: bool):
    chain = chain.lower()
    parents = DECODEX_DIR.joinpath(chain)

    if chain == "ethereum":
        download_github_file(
            save_path=str(parents.joinpath("tags.json")),
            org="brianleect",
            repo="etherscan-labels",
            branch="main",
            path="data/etherscan/combined/combinedAllLabels.json",
            is_lfs=False,
            verify_ssl=verify_ssl,
            use_tempfile=False,
        )
        download_github_file(
            save_path=str(parents.joinpath("signatures.csv")),
            org="Solratic",
            repo="function-signature-registry",
            branch="main",
            path="data/ethereum/func_sign.csv.gz",
            is_lfs=True,
            verify_ssl=verify_ssl,
            use_tempfile=True,
        )
    else:
        raise ValueError(f"Chain {chain} is not yet supported.")


@cli.command(help="Remove downloaded tags and signatures for a chain")
@click.argument("chain", default="ethereum", type=click.Choice(["ethereum"]))
def clean(chain: str):
    chain = chain.lower()
    dir = DECODEX_DIR.joinpath(chain)
    if dir.exists():
        for f in dir.iterdir():
            f.unlink()
        dir.rmdir()
    else:
        print(f"{chain} not found")


@cli.command(help="Explain the transaction by the given hash")
@click.option("--txhash", type=str, help="Hash of the transaction", default=None)
@click.option("--from-addr", type=str, help="Address of the sender", default=None)
@click.option("--to-addr", type=str, help="Address of the receiver", default=None)
@click.option("--value", type=float, help="Value in ether", default=None)
@click.option("--input-data", type=str, help="Input data of the transaction", default="0x")
@click.option("--block", type=str, help="Block number of the transaction", default="latest")
@click.option("--chain", "-c", type=click.Choice(["ethereum"]), default="ethereum", help="Chain to use for decoding")
@click.option("--provider-uri", "-p", type=str, default=os.getenv("WEB3_PROVIDER_URI", "http://localhost:8545"))
@click.option("--verbose", "-v", is_flag=True, help="Verbose mode", default=False)
def explain(
    txhash: str,
    from_addr: str,
    to_addr: str,
    value: int,
    input_data: str,
    block: Union[Literal["latest"], int],
    chain: str,
    provider_uri: str,
    verbose: bool,
):
    translator = Translator(provider_uri=provider_uri, chain=chain, verbose=verbose)
    if txhash is None:
        # If txhash is not provided, then from_addr, to_addr, value, input_data must be provided to simulate a transaction
        assert (
            from_addr is not None and to_addr is not None and value is not None and input_data is not None
        ), "Either txhash or [from_addr, to_addr, value, input_data] must be provided"
        assert (
            from_addr.startswith("0x") and to_addr.startswith("0x") and input_data.startswith("0x")
        ), "from_addr, to_addr, input_data must be hex string"
        assert value >= 0, "value must be non-negative"
        value = int(value * 1e18)
        block = int(block) if block != "latest" else block
        tagged_tx = translator.simulate(
            from_address=from_addr,
            to_address=to_addr,
            value=value,
            data=input_data,
            block=block,
        )
    else:
        # If txhash is provided
        tagged_tx = translator.translate(txhash=txhash)

    tmpl = Template(
        """\
Transaction: {{txhash}}
Blocktime: {{blocktime}}
From: {{from_addr}}
To: {{to_addr}} {% if contract_created %} (Contract Created: {{contract_created}}){% endif %}
Value: {{value}}
GasUsed: {{gas_used}}
Gas Price: {{gas_price}}
Status: {{status}}{% if reason %} ({{reason}}){% endif %}
{% if method -%}
Method: {{method}}
{% endif -%}
{% if actions -%}

Actions
------
{{actions}}
{% endif -%}
{% if balance_change -%}

Balance Changes
--------------
{{balance_change}}
{% endif -%}
"""
    )
    txhash = tagged_tx["txhash"]
    blocktime = fmt_blktime(tagged_tx["block_time"])
    from_addr = fmt_addr(tagged_tx["from"])
    to_addr = fmt_addr(tagged_tx["to"]) if tagged_tx["to"] else None
    contract_created = fmt_addr(tagged_tx["contract_created"]) if tagged_tx["contract_created"] else None
    value = fmt_value(tagged_tx["value"])
    gas_used = f'{tagged_tx["gas_used"]}'
    gas_price = fmt_gas(tagged_tx["gas_price"])
    status = fmt_status(tagged_tx["status"])
    reason = tagged_tx["reason"]

    action_str = "\n".join(f"- {a}" for a in tagged_tx["actions"])
    indented_actions = indent(action_str, " " * 4)

    render = tmpl.render(
        txhash=txhash,
        blocktime=blocktime,
        from_addr=from_addr,
        to_addr=to_addr,
        contract_created=contract_created,
        value=value,
        gas_used=gas_used,
        gas_price=gas_price,
        status=status,
        reason=reason,
        method=tagged_tx["method"],
        actions=indented_actions,
    )

    for acc in tagged_tx["balance_change"]:
        render += "\n"
        render += "Account: " + fmt_addr(acc["address"], truncate=False)
        render += "\n"
        table_data = []
        for asset in acc["assets"]:
            table_data.append(
                [
                    fmt_addr(asset["asset"], truncate=False),
                    asset["balance_change"],
                ]
            )

        render += tabulate(table_data, headers=["Asset", "Balance Change"], tablefmt="grid")
        render += "\n"

    indented_render = indent(render, "  ")  # Indent entire tmpl by 2 spaces
    print(indented_render)


if __name__ == "__main__":
    cli()



from typing import Dict


def get_github_api_url(org: str, repo: str, branch: str, path: str) -> str:
    return f"https://api.github.com/repos/{org}/{repo}/contents/{path}?ref={branch}"


def get_github_rawfile_url(org: str, repo: str, branch: str, path: str) -> str:
    return f"https://raw.githubusercontent.com/{org}/{repo}/{branch}/{path}"


def get_github_url(org: str, repo: str, branch: str, path: str, use_api: bool) -> str:
    if use_api:
        return get_github_api_url(org=org, repo=repo, branch=branch, path=path)
    else:
        return get_github_rawfile_url(org=org, repo=repo, branch=branch, path=path)



from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

from eth_abi import decode as decode_abi
from eth_utils.abi import collapse_if_tuple


def eth_decode_input(abi: Dict, data: str) -> Tuple[str, Dict]:
    """
    Decodes Ethereum input given the ABI and data.

    Parameters
    ----------
    abi : Dict
        The ABI of the function.
    data : str
        The data associated with the function.

    Returns
    -------
    Tuple[str, Dict]
        The function signature and the parameters decoded from the input.

    """

    # Ensure ABI is a valid function
    if "name" not in abi or abi.get("type") != "function":
        return "{}", {}

    # Separate inputs into indexed and non-indexed
    inputs = abi.get("inputs", [])

    func_signature = [collapse_if_tuple(inp) for inp in inputs]
    text_signature = "{}({})".format(abi.get("name", ""), ",".join(func_signature))

    # Decode values from data
    values = decode_abi(func_signature, bytes(bytearray.fromhex(data[10:])))

    # Fill parameters with values
    params = {}
    for idx, val in enumerate(values):
        params.update(_process_abi_tuple(inputs[idx], val))

    return text_signature, params


def _process_abi_tuple(abi: Dict, value: Any) -> Dict:
    """
    Process ABI (Application Binary Interface) and value based on type information.

    Parameters
    ----------
    abi : Dict
        The ABI information as a dictionary, including 'type' and 'name' keys.
    value : Any
        The value to be processed, the type of which depends on the ABI information.

    Returns
    -------
    Dict
        A dictionary containing the processed value(s), with the key as the name from the ABI.

    Notes
    -----
    The function can handle types like 'byte', 'tuple' and their array forms.
    """
    abi_type: str = abi["type"]
    abi_name: str = abi["name"]

    if abi_type.startswith("byte"):
        return _process_byte_type(abi_type, abi_name, value)

    if abi_type.startswith("tuple"):
        return _process_tuple_type(abi, abi_name, value)

    return {abi_name: value}


def _process_byte_type(abi_type: str, abi_name: str, value: Union[bytes, List[bytes]]) -> Dict:
    if abi_type.endswith("]"):
        hex_values = [element.hex() for element in value]
        processed_values = []

        for hex_value in hex_values:
            chunks = _divide_into_chunks(hex_value, 64)
            processed_values.extend(chunks)

        return {abi_name: processed_values}

    return {abi_name: value.hex()}


def _process_tuple_type(abi: Dict, abi_name: str, value: Any) -> Dict:
    is_array = len(abi["type"]) > len("tuple")
    sub_abis = abi["components"]

    if not is_array:
        processed_value = _process_single_tuple(sub_abis, value)
        return {abi_name: processed_value}

    processed_values = [_process_single_tuple(sub_abis, sub_value) for sub_value in value]
    return {abi_name: processed_values}


def _process_single_tuple(sub_abis: Dict, value: Any) -> Dict:
    processed_value = {}
    for index, sub_abi in enumerate(sub_abis):
        processed_value.update(_process_abi_tuple(sub_abi, value[index]))
    return processed_value


def _divide_into_chunks(data: str, chunk_size: int) -> list:
    """
    Divide the data into chunks of a given size. The last chunk is zero-padded if needed.

    Parameters:
        data (str): The string data to be chunked.
        chunk_size (int): The size of each chunk.

    Returns:
        list: The list of chunks.
    """
    chunks = [data[i : i + chunk_size] for i in range(0, len(data), chunk_size)]

    if len(chunks[-1]) < chunk_size:
        padding_size = chunk_size - len(chunks[-1])
        chunks[-1] += "0" * padding_size

    return chunks


def eth_decode_log(event_abi: Dict, topics: List[str], data: str) -> Tuple[str, Dict]:
    """
    Decodes Ethereum log given the event ABI, topics, and data.
    """

    # Ensure ABI is a valid event
    if "name" not in event_abi or event_abi.get("type") != "event":
        return "{}", {}

    inputs: List = event_abi.get("inputs", [])

    # Separate indexed and non-indexed inputs
    indexed_types, non_indexed_types = [], []
    indexed_idx, non_indexed_idx = [], []
    for idx, inp in enumerate(inputs):
        if inp.get("indexed"):
            indexed_types.append(collapse_if_tuple(inp))
            indexed_idx.append(idx)
        else:
            non_indexed_types.append(collapse_if_tuple(inp))
            non_indexed_idx.append(idx)

    indexed_values = decode_abi(indexed_types, bytes(bytearray.fromhex("".join(t[2:] for t in topics[1:]))))
    non_indexed_values = decode_abi(non_indexed_types, bytes(bytearray.fromhex(data[2:])))

    params = {}
    for idx, value in zip(indexed_idx, indexed_values):
        single_param = _process_abi_tuple(inputs[idx], value)
        params.update(single_param)
        params.update({f"__idx_{idx}": list(single_param.values())[0]})
    for idx, value in zip(non_indexed_idx, non_indexed_values):
        single_param = _process_abi_tuple(inputs[idx], value)
        params.update(single_param)
        params.update({f"__idx_{idx}": list(single_param.values())[0]})

    func_signature = [collapse_if_tuple(inp) for inp in inputs]
    text_signature = "{}({})".format(event_abi.get("name", ""), ",".join(func_signature))

    return text_signature, params



import json
import os
from abc import ABC
from abc import abstractmethod
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Generator
from typing import Literal
from typing import Optional
from typing import Sequence
from typing import Union

from decodex.constant import DECODEX_DIR
from decodex.type import ERC20Compatible
from decodex.type import TaggedAddr


class AddrTagger(ABC):
    @abstractmethod
    def lazy_tag(
        self, address: Sequence[Optional[Union[str, Dict]]]
    ) -> Generator[Union[TaggedAddr, ERC20Compatible], None, None]:
        """
        Tag addresses.
        """
        raise NotImplementedError

    @abstractmethod
    def __call__(
        self, address: Union[Sequence[Optional[Union[str, Dict]]], Optional[str]]
    ) -> Sequence[Optional[Union[TaggedAddr, ERC20Compatible]]]:
        """
        Tag one or more addresses.
        """
        raise NotImplementedError


class SyncAddrTagger(AddrTagger):
    """
    Tag addresses synchronously, one by one.
    """

    def __call__(
        self, address: Union[Sequence[Optional[Union[str, Dict]]], Optional[str]]
    ) -> Sequence[Optional[Union[TaggedAddr, ERC20Compatible]]]:
        if address is None:
            return [None]
        if isinstance(address, str):
            address = [address]
        return list(self.lazy_tag(address))


class JSONAddrTagger(SyncAddrTagger):
    """
    A class for tagging addresses using data from a JSON file.

    This class extends the functionality of SyncAddrTagger to provide address tagging using information
    stored in a JSON file. The JSON file should contain address-tag pairs, where each address is associated
    with metadata like name and labels.

    Parameters
    ----------
        path (str): The path to the JSON file containing address-tag data. Defaults to `~/.decodex/tags.json`

    Raises
    ------
        AssertionError: If the provided path does not correspond to a valid file or is not a JSON file.


    Example
    -------
    To use the JSONAddrTagger, create an instance with the path to a valid JSON file and then call
    the instance with a single address or a list of addresses to obtain tagging information.

    ```python
    tagger = JSONAddrTagger("address_tags.json")
    tagged_addresses = tagger(["0x22ff777ef6fe0690f1f74c6758126909653ad56a"])
    for addr in tagged_addresses:
        print(addr)
    ```
    """

    def __init__(self, path: str = None, chain: str = "ethereum") -> None:
        if path is None:
            path = str(DECODEX_DIR.joinpath(chain, "tags.json"))
        assert os.path.isfile(path), f"Path {path} is not a file."
        assert path.endswith(".json"), f"Path {path} is not a JSON file."

        with open(path, "r") as file:
            self._addr_tags: Dict[str, Any] = json.load(file)

    def lazy_tag(
        self,
        address: Sequence[Optional[Union[str, Dict]]],
    ) -> Generator[Optional[Union[TaggedAddr, ERC20Compatible]], None, None]:
        for addr in address:
            if addr is None:
                yield None
                continue
            if isinstance(addr, dict):
                _relying_addr: Optional[str] = addr.get("address", None)
                if _relying_addr is None or not isinstance(_relying_addr, str):
                    yield None
                    continue
                tag: Dict[str, Any] = self._addr_tags.get(_relying_addr.lower(), {})
                extra_fields = {
                    "name": tag.get("name", ""),
                    "labels": tag.get("labels", []),
                }
                yield {**addr, **extra_fields}
            else:
                tag: Dict[str, Any] = self._addr_tags.get(addr.lower(), {})
                yield {
                    "address": addr,
                    "name": tag.get("name", ""),
                    "labels": tag.get("labels", []),
                }


class BatchAddrTagger(AddrTagger):
    def __init__(self) -> None:
        super().__init__()

    def __call__(
        self, address: Sequence[Optional[str]] | Optional[str]
    ) -> Sequence[Union[TaggedAddr, ERC20Compatible]]:
        raise NotImplementedError


class TaggerFactory:
    @staticmethod
    def create(
        tagger_type: Literal["json", "sql"],
        uri: str = None,
        chain: str = "ethereum",
    ) -> AddrTagger:
        if tagger_type == "json":
            return JSONAddrTagger(path=uri, chain=chain)
        else:
            raise ValueError(f"Unknown tagger type: {tagger_type}")



from .fs import DECODEX_DIR


NULL_ADDRESS_0x0 = "0x0000000000000000000000000000000000000000"
NULL_ADDRESS_0xF = "0xffffffffffffffffffffffffffffffffffffffff"

__all__ = [
    "NULL_ADDRESS_0x0",
    "NULL_ADDRESS_0xF"
    #
    "DECODEX_DIR",
]



from typing import Any
from typing import List
from typing import Literal
from typing import Optional
from typing import TypedDict
from typing import Union

from .tx_type import Log


RawTraceCall = TypedDict(
    "RawTraceCall",
    {
        "from": str,
        "gas": str,
        "gasUsed": str,
        "input": str,
        "logs": List[Log],
        "output": str,
        "to": str,
        "type": Literal["CALL", "CALLCODE", "DELEGATECALL", "STATICCALL"],
        "value": str,
        "calls": List["RawTraceCall"],
    },
    total=False,
)

RawTraceCallResult = TypedDict(
    "RawTraceCallResult",
    {
        "from": str,
        "gas": str,
        "gasUsed": str,
        "input": str,
        "to": str,
        "type": Literal["CALL", "CALLCODE", "DELEGATECALL", "STATICCALL"],
        "calls": List[RawTraceCall],
    },
    total=False,
)


class RPCError(TypedDict, total=False):
    code: int
    message: str
    gasFeeCap: int
    baseFee: int


class RawTraceCallResponse(TypedDict, total=False):
    id: Union[str, int]
    jsonrpc: str
    result: RawTraceCallResult
    error: Optional[RPCError]



import json
import traceback
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from logging import Logger
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Literal
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

import pytz
from multicall import Call
from multicall import Multicall
from web3 import Web3

from decodex.constant import NULL_ADDRESS_0x0
from decodex.constant import NULL_ADDRESS_0xF
from decodex.convert.address import AddrTagger
from decodex.convert.address import TaggerFactory
from decodex.convert.signature import SignatureFactory
from decodex.convert.signature import SignatureLookUp
from decodex.decode import eth_decode_input
from decodex.decode import eth_decode_log
from decodex.search import SearcherFactory
from decodex.translate.events import AAVEV2Events
from decodex.translate.events import AAVEV3Events
from decodex.translate.events import BancorEV3Events
from decodex.translate.events import CompoundV3Events
from decodex.translate.events import CurveV2Events
from decodex.translate.events import ERC20Events
from decodex.translate.events import UniswapV2Events
from decodex.translate.events import UniswapV3Events
from decodex.type import AccountBalanceChanged
from decodex.type import Action
from decodex.type import AssetBalanceChanged
from decodex.type import ContractCreation
from decodex.type import EventHandleFunc
from decodex.type import TaggedAddr
from decodex.type import TaggedTx
from decodex.type import TransferAction
from decodex.type import Tx
from decodex.type import UTF8Message
from decodex.utils import parse_ether
from decodex.utils import parse_gwei
from decodex.utils import parse_utf8


class Translator:
    evt_opts: Dict[str, Any] = {
        "uniswapv2": UniswapV2Events,
        "uniswapv3": UniswapV3Events,
        "bancorv3": BancorEV3Events,
        "curvev2": CurveV2Events,
        "aavev2": AAVEV2Events,
        "aavev3": AAVEV3Events,
        "compoundv3": CompoundV3Events,
    }

    def __init__(
        self,
        provider_uri: str,
        chain: str = "ethereum",
        tagger: AddrTagger = "json",
        sig_lookup: SignatureLookUp = "csv",
        defis: Union[Iterable[str], Literal["all"]] = "all",
        verbose: bool = False,
        logger: Logger = None,
        skip_install: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        provider_uri : str
            URI of the Ethereum http provider
        tagger : AddrTagger, optional
            Address tagger or the `tagger_types` in TaggerFactory, default is "json"
        sig_lookup : SignatureLookUp, optional
            Signature lookup or the `sig_lookup_types` in SignatureFactory, default is "csv"
        defis : Union[Iterable[str], Literal["all"]], optional
            List of defi protocols to decode or "all" to decode all supported protocols, default is "all"
            You can get the list of supported protocols by calling `Translator.supported_defis()`
        verbose : bool, optional
            Whether to print error messages when decoding logs, default is False
        logger : Logger, optional
            Logger to log error messages, default is None
        """

        self.tagger = TaggerFactory.create(tagger) if isinstance(tagger, str) else tagger

        self.sig_lookup = (
            SignatureFactory.create(fmt=sig_lookup, chain=chain) if isinstance(sig_lookup, str) else sig_lookup
        )

        self.chain = chain
        self.searcher = SearcherFactory.create("web3", uri=provider_uri)
        self.mc = Multicall(provider_uri, logger=logger)
        self.hdlrs: Dict[str, EventHandleFunc] = {}
        self.web3 = Web3(Web3.HTTPProvider(provider_uri))
        self.__register__(self.evt_opts.keys() if defis == "all" else defis)

        self.verbose = verbose
        if logger is None and verbose:
            self.logger = Logger(name=self.__class__.__name__)

        if not skip_install:
            self.install()

    def install(self):
        from decodex import installer
        from decodex import constant

        tags_path = constant.DECODEX_DIR.joinpath(self.chain, "tags.json")
        if not tags_path.exists():
            installer.download_github_file(
                save_path=str(tags_path),
                org="brianleect",
                repo="etherscan-labels",
                branch="main",
                path="data/etherscan/combined/combinedAllLabels.json",
                is_lfs=False,
                verify_ssl=False,
                use_tempfile=False,
            )

        signature_path = constant.DECODEX_DIR.joinpath(self.chain, "signatures.csv")
        if not signature_path.exists():
            installer.download_github_file(
                save_path=str(signature_path),
                org="Solratic",
                repo="function-signature-registry",
                branch="main",
                path="data/ethereum/func_sign.csv.gz",
                is_lfs=True,
                verify_ssl=False,
                use_tempfile=True,
            )

    def translate(self, txhash: str, *, max_workers: int = 10) -> TaggedTx:
        tx: Tx = self.searcher.get_tx(txhash)
        return self._process_tx(tx, max_workers=max_workers)

    def simulate(
        self,
        from_address: str,
        to_address: str,
        value: int,
        data: str,
        block: Union[int, Literal["latest"]] = "latest",
        *,
        gas: Union[int, Literal["auto"]] = "auto",
        gas_price: Union[int, Literal["auto"]] = "auto",
        timeout: int = 120,
        max_workers: int = 10,
    ) -> TaggedTx:
        from_address = Web3.to_checksum_address(from_address.lower())
        to_address = Web3.to_checksum_address(to_address.lower())
        simulated_tx = self.searcher.simluate_tx(
            from_address=from_address,
            to_address=to_address,
            value=value,
            data=data,
            block=block,
            gas=gas,
            gas_price=gas_price,
            timeout=timeout,
        )
        return self._process_tx(simulated_tx, max_workers=max_workers)

    @classmethod
    def supported_defis(cls) -> List[str]:
        return list(cls.evt_opts.keys())

    def __register__(self, defis: Iterable[str]) -> None:
        # ERC20Events must be registered
        opts = self.evt_opts
        opts["erc20"] = ERC20Events

        for defi in defis:
            assert defi in opts, f"defi protocol {defi} is not yet supported"
            cls = opts.get(defi)(self.mc, self.tagger, self.sig_lookup)
            for attr in dir(cls):
                if attr.startswith("_"):
                    continue
                handle_func = getattr(cls, attr)
                if not callable(handle_func):
                    continue
                text_sig, decoder = handle_func()
                byte_sig = Web3.keccak(text=text_sig).hex()
                self.hdlrs[byte_sig] = decoder

    def _decode_log(self, log: Dict[str, Any]) -> Optional[Action]:
        topics = log.get("topics", [])
        if len(topics) == 0:
            raise ValueError("Log topics is empty")
        handler = self.hdlrs.get(topics[0], None)
        if handler is None:
            return None
        abi_textsign_list = self.sig_lookup(topics[0])
        for abi, _ in abi_textsign_list:
            abi = json.loads(abi)
            try:
                _, params = eth_decode_log(abi, topics, log.get("data", "0x"))
                result = handler({"address": log["address"], "params": params})
                return result
            except Exception as e:
                if self.verbose:
                    traceback.print_exc()
                    self.logger.error(f"Error when decoding log {log} with error {e}")
        return None

    def _decode_input(self, data: str) -> str:
        if not data or len(data) < 10:
            return ""
        func_selector = data[:10]
        candidates = self.sig_lookup(func_selector)
        for abi, _ in candidates:
            try:
                abi = json.loads(abi)
                _, params = eth_decode_input(abi, data)
                func = abi["name"]
                # Format the output string into funcA(arg1=a, arg2=b)
                if len(params) > 0:
                    func += (
                        "("
                        + ", ".join(
                            [
                                f"{k}={v}" if abi["inputs"][_]["type"] not in {"bytes[]"} else f"{k}=..."
                                for _, (k, v) in enumerate(params.items())
                            ]
                        )
                        + ")"
                    )
                return func
            except Exception as e:
                if self.verbose:
                    traceback.print_exc()
                    self.logger.error(f"Error when decoding input {data} with error {e}")
                continue
        return func_selector

    def _get_erc20_balabce(
        self,
        addr_token_pairs: Iterable[Tuple[str, str]],
        blk_num: int,
    ) -> Dict[str, float]:
        """
        Get the balance of a list of (address, ERC20 token) pairs, and return a dict of str((addr,token)) -> balance.
        """

        # Get all decimals
        decimal_calls: List[Call] = []
        involved_tokens: Set[str] = set(token for _, token in addr_token_pairs)
        for token in involved_tokens:
            decimal_calls.append(
                Call(
                    target=token,
                    function="decimals()(uint8)",
                    request_id=token,
                    block_id=blk_num,
                )
            )
        decimals: Dict[str, int] = self.mc.agg(decimal_calls, as_dict=True)
        # Get all balances
        calls: List[Call] = []
        for addr, token in addr_token_pairs:
            calls.append(
                Call(
                    target=token,
                    function="balanceOf(address)(uint256)",
                    args=(addr,),
                    request_id=str((addr, token)),
                    block_id=blk_num,
                )
            )
        balances: Dict[str, int] = self.mc.agg(calls, as_dict=True)
        return {
            str((addr, token)): balances[str((addr, token))] / 10 ** decimals[token] for addr, token in addr_token_pairs
        }

    def _get_eth_balance(
        self,
        addrs: Iterable[str],
        blk_num: Union[int, Literal["latest"]],
        max_workers: int,
    ) -> Dict[str, int]:
        """
        Parameters
        ----------
        addrs : Iterable[str]
            List of addresses, the address must be in checksum format
        blk_num : int
            Block number

        Returns
        -------
        Dict[str, int]
            A dict of address -> balance
            the return address (key) is in lowercase format
        """

        def proxy(addr: str) -> int:
            return self.web3.eth.get_balance(addr, block_identifier=blk_num)

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            balances = executor.map(proxy, addrs)
        return dict(zip([addr.lower() for addr in addrs], balances))

    def _process_transfer(
        self,
        actions: List[TransferAction],
        eth_balance_changes: Dict[str, Dict[Literal["ETH", "Gas Fee"], int]],
    ) -> List[AccountBalanceChanged]:
        tagged_mapping: Dict[str, TaggedAddr] = {}
        addr_token_pairs = set()

        for action in actions:
            sender = action.sender["address"]
            receiver = action.receiver["address"]
            token = action.token["address"]

            tagged_mapping[sender] = action.sender
            tagged_mapping[receiver] = action.receiver
            tagged_mapping[token] = action.token

            addr_token_pairs.add((sender, token))
            addr_token_pairs.add((receiver, token))

        balance_changed: Dict[str, Dict[str, AssetBalanceChanged]] = {}
        for account, token in addr_token_pairs:
            if account not in balance_changed:
                balance_changed[account] = {}

            balance_changed[account][token] = {
                "asset": tagged_mapping[token],
                "balance_change": 0.0,
            }

        for action in actions:
            sender = action.sender["address"]
            receiver = action.receiver["address"]
            token = action.token["address"]
            amount = action.amount

            balance_changed[sender][token]["balance_change"] -= amount
            balance_changed[receiver][token]["balance_change"] += amount

        extra_account = set([e.lower() for e in eth_balance_changes.keys()]) - set(tagged_mapping.keys())
        eth_tagged_mapping = zip(extra_account, self.tagger(extra_account))
        tagged_mapping.update(eth_tagged_mapping)
        for account, changes in eth_balance_changes.items():
            account = account.lower()

            if balance_changed.get(account) is None:
                balance_changed[account] = {}

            if changes.get("ETH", 0):
                balance_changed[account]["ETH"] = {
                    "asset": {"address": NULL_ADDRESS_0x0, "name": "Platform (Ether)", "labels": []},
                    "balance_change": changes["ETH"] / 10**18,
                }

            if changes.get("Gas Fee", 0):
                balance_changed[account]["Gas Fee"] = {
                    "asset": {"address": NULL_ADDRESS_0xF, "name": "Platform (Gas Fee)", "labels": []},
                    "balance_change": changes["Gas Fee"] / 10**18,
                }

        account_balance_changed_list: List[AccountBalanceChanged] = [
            {
                "address": tagged_mapping[account],
                "assets": [asset for asset in assets_dict.values() if asset["balance_change"] != 0],
            }
            for account, assets_dict in balance_changed.items()
        ]

        return account_balance_changed_list

    def _process_tx(self, tx: Tx, max_workers: int) -> TaggedTx:
        # Decode the events
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            actions = executor.map(self._decode_log, tx["logs"])
        actions = [x for x in actions if x is not None]

        # Split actions into transfers and others
        transfers: List[TransferAction] = []
        others: List[Action] = []
        for action in actions:
            if isinstance(action, TransferAction):
                transfers.append(action)
            else:
                others.append(action)

        bal_changed = self._process_transfer(
            actions=transfers,
            eth_balance_changes=tx["eth_balance_changes"],
        )

        # Get the time of the block
        blk_time = datetime.fromtimestamp(tx["block_timestamp"], tz=pytz.utc)

        # Tag the addresses
        tx_from = self.tagger(tx["from"])[0]
        tx_to = self.tagger(tx["to"])[0] if tx["to"] else None
        tx_contract_created = self.tagger(tx["contract_created"])[0] if tx["contract_created"] else None

        # Get the method
        method = self._decode_input(tx["input"])

        # If there are other actions, then ignore transfers
        if len(others) > 0:
            actions = others
        elif len(transfers) > 0:
            actions = transfers
        elif tx.get("contract_created"):
            actions = [ContractCreation(deployer=tx_from, contract=tx_contract_created)]
        elif len(tx["input"]) > 0 and tx["input"] != "0x":
            msg = parse_utf8(tx["input"])
            if msg:
                actions = [UTF8Message(tx_from, tx_to, msg)]
        else:
            actions = []

        return {
            "txhash": tx["txhash"],
            "from": tx_from,
            "to": tx_to,
            "contract_created": tx_contract_created,
            "block_number": tx["block_number"],
            "block_time": blk_time,
            "value": parse_ether(tx["value"]),
            "gas_used": tx["gas_used"],
            "gas_price": parse_gwei(tx["gas_price"]),
            "input": tx["input"],
            "status": tx["status"],
            "reason": tx["reason"],
            "actions": actions,
            "method": method,
            "balance_change": bal_changed,
        }



import json
import os
from typing import Callable
from typing import Dict
from typing import Generator
from typing import Literal
from typing import Tuple

import pandas as pd

from decodex.constant import DECODEX_DIR


class SignatureLookUp:
    def __init__(self) -> None:
        pass

    def __call__(self, byte_sign: str) -> Generator[Tuple[str, str], None, None]:
        """
        Search for the signature in the database and return the corresponding abi and text_signature.

        Parameters
        ----------
        byte_sign : str
            The signature in hex string, 0x prefixed.

        Returns
        -------
        Tuple[Tuple[str, str]]:
            The list of (abi, signature) corresponding to the signature.

        """
        raise NotImplementedError


class CSVSignatureLookUp(SignatureLookUp):
    def __init__(self, uri: str = None, chain: str = "ethereum") -> None:
        super().__init__()
        if uri is None:
            uri = DECODEX_DIR.joinpath(chain, "signatures.csv")
            if not os.path.isfile(uri):
                raise ValueError(f"Signature lookup file {uri} does not exist")
        self.df = pd.read_csv(uri)
        self.__validate__()

    def __validate__(self):
        if self.df.empty:
            raise ValueError("Signature lookup file is empty")

        cols = set(self.df.columns)
        expected_cols = {"byte_sign", "abi", "text_sign", "score"}
        if cols != expected_cols:
            raise ValueError(
                f"Signature lookup file is not valid, expected columns: {', '.join(expected_cols)}, but got: {', '.join(cols)}"
            )

    def __call__(self, byte_sign: str) -> Generator[Tuple[str, str], None, None]:
        rows = self.df[self.df["byte_sign"] == byte_sign]
        if rows.empty:
            return {}, ""

        # If there are multiple rows, return the one with the highest score
        rows = rows.sort_values(by="score", ascending=False)

        # Return all rows
        for _, row in rows.iterrows():
            yield row["abi"], row["text_sign"]


class SignatureFactory:
    """
    A factory class to instantiate various types of SignatureLookUp classes based on the given format.

    Methods
    -------
    create(fmt: Literal["csv"], uri: str = None) -> SignatureLookUp:
        Creates and returns an instance of a SignatureLookUp subclass based on the specified format.

    Example
    -------
    >>> factory = SignatureFactory()
    >>> csv_lookup = factory.create("csv", uri="path/to/signatures.csv")

    Parameters for static methods
    ------------------------------
    fmt : Literal["csv"]
        The format of the SignatureLookUp to instantiate. Currently, only "csv" is supported.
    uri : str, optional
        The URI where the signature lookup file can be found. Defaults to None, which may use a defaul value in their constructor.
        For example, CSVSignatureLookUp uses "~/.decodex/signatures.csv" as the default value.

    Returns
    -------
    SignatureLookUp
        An instance of the corresponding SignatureLookUp subclass.

    Raises
    ------
    ValueError
        If the given `fmt` is not supported.
    """

    @staticmethod
    def create(
        fmt: Literal["csv"],
        uri: str = None,
        chain: str = "ethereum",
    ) -> SignatureLookUp:
        if fmt == "csv":
            return CSVSignatureLookUp(uri=uri, chain=chain)
        else:
            raise ValueError(f"Signature lookup format {fmt} is not supported")



import gzip
import hashlib
import json
from abc import abstractstaticmethod
from io import BytesIO
from pathlib import Path

import requests

from .utils import get_github_url


class BaseBeforeCallback:
    @abstractstaticmethod
    def before_download(save_path: str, **kwargs) -> str:
        pass


class BasePostCallback:
    @abstractstaticmethod
    def post_download(src_path: str, save_path: str, **kwargs):
        pass


class GithubRawBeforeCallback(BaseBeforeCallback):
    """
    kwargs : Dict
        The arguments to pass to get_github_url.
        - org : str
            The name of the organization or user.
        - repo : str
            The name of the repository.
        - branch : str
            The name of the branch.
        - path : str
            The path to the file in the repository.
    """

    def before_download(save_path: str, **kwargs) -> str:
        url = get_github_url(use_api=False, **kwargs)
        return url


class GithubLFSBeforeCallback(BaseBeforeCallback):
    @staticmethod
    def before_download(save_path: str, **kwargs) -> str:
        """
        Get spec from Github API then check the checksum of the file to download.

        Parameters
        ----------
        save_path : str
            The path to save the file to.
        kwargs : Dict
            The arguments to pass to get_github_lfs_url.
            - org : str
                The name of the organization or user.
            - repo : str
                The name of the repository.
            - branch : str
                The name of the branch.
            - path : str
                The path to the file in the repository.
        Returns
        -------
        str
            The modified URL.

        Raises
        ------
        FileExistsError
            If the file already exists and is up to date.
        """
        spec_url = get_github_url(use_api=True, **kwargs)

        spec = requests.request(method="GET", url=spec_url)

        spec.raise_for_status()
        spec = json.loads(spec.text)
        download_url = spec["download_url"]
        checksum = spec["sha"]

        # Check if the file already exists
        path = Path(save_path)
        if path.exists() and path.with_suffix(".checksum").exists():
            # Check if the checksum is the same
            with path.with_suffix(".checksum").open("r") as cf:
                if cf.read() == checksum:
                    raise FileExistsError("File already exists and is up to date")
        # Save the checksum
        path.with_suffix(".checksum").write_text(checksum)
        return download_url


class GzipPostCallback(BasePostCallback):
    @staticmethod
    def post_download(src_path: str, save_path: str, **config):
        """
        Decompress a gzip file.

        Parameters
        ----------
        src_path : str
            The path to the file to decompress.
        save_path : str
            The path to save the decompressed file to.
        """
        in_path = Path(src_path)
        out_path = Path(save_path)
        with in_path.open("rb") as f, out_path.open("wb") as g:
            compressed_data = BytesIO(f.read())
            decompressed_file = gzip.GzipFile(fileobj=compressed_data)
            g.write(decompressed_file.read())


class ChecksumPostCallback(BasePostCallback):
    @staticmethod
    def post_download(src_path: str, save_path: str, **config):
        """
        Calculate the checksum of a file and save it to {save_path}.checksum.

        Parameters
        ----------
        src_path : str
            The path to the file to check.
        save_path : str
            The path to save the file to.
        """
        # Define the path objects
        checksum_path = Path(save_path).with_suffix(".checksum")

        # Calculate the checksum of the file
        hash_algorithm = config.get("hash_algorithm", config.get("algorithm", "sha256"))
        hash_object = hashlib.new(hash_algorithm)

        with Path(src_path).open("rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_object.update(chunk)

        checksum = hash_object.hexdigest()

        # Save the checksum to the specified path
        with checksum_path.open("w") as g:
            g.write(checksum)



from .fmt import fmt_addr
from .fmt import fmt_blktime
from .fmt import fmt_gas
from .fmt import fmt_status
from .fmt import fmt_value
from .utils import parse_ether
from .utils import parse_gwei
from .utils import parse_unit
from .utils import parse_utf8
from .utils import trunc_addr

__all__ = [
    "trunc_addr",
    "parse_ether",
    "parse_gwei",
    "parse_unit",
    "parse_utf8",
    "fmt_blktime",
    "fmt_addr",
    "fmt_gas",
    "fmt_value",
    "fmt_status",
    "fmt_actions",
]



from decimal import Decimal
from typing import Optional
from typing import Union


def trunc_addr(addr: str, offset: int = 4) -> str:
    return addr[: offset + 2] + "..." + addr[-offset:]


def parse_ether(wei: int) -> float:
    return Decimal(wei) / 10**18


def parse_utf8(hex: str) -> Optional[str]:
    """
    Parameters
    ----------
    hex : str
        Hex string to be decoded. 0x prefixed.
    """
    try:
        byte_data = bytes.fromhex(hex[2:])
        utf8_str = byte_data.decode("utf-8")
        return utf8_str
    except Exception as e:
        return None


def parse_gwei(wei: int) -> float:
    return Decimal(wei) / 10**9


def parse_unit(wei: Union[int, str], unit: str) -> str:
    if unit == "eth":
        return parse_ether(wei)
    elif unit == "gwei":
        return parse_gwei(wei)
    else:
        return f"{wei} {unit}"



from . import constant
from . import convert
from . import decode
from . import installer
from . import search
from . import translate
from . import type
from . import utils


__all__ = [
    "convert",
    "decode",
    "search",
    "translate",
    "type",
    "utils",
    "installer",
]



from datetime import datetime
from typing import Dict
from typing import Optional

from .utils import trunc_addr


def fmt_blktime(blktime: datetime) -> str:
    return blktime.strftime("%Y-%m-%d %H:%M:%S") + " UTC"


def fmt_addr(ta: Dict, truncate: bool = True) -> str:
    addr = ta["address"]
    name = ta["name"]
    labels = ", ".join(l for l in ta["labels"] if l)

    if truncate:
        rtn = name if name else trunc_addr(addr)
    else:
        rtn = f"{addr} ({name})" if name else addr
    if labels:
        rtn += f" [{labels}]"
    return rtn


def fmt_gas(gas_price: int) -> str:
    return f"{gas_price} Gwei"


def fmt_value(value: int) -> str:
    return f"{value} Ether"


def fmt_status(status: int) -> str:
    return "Success" if status == 1 else "Failed"



from abc import abstractmethod
from dataclasses import asdict
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional

from .base import Action
from .tx_type import ERC20Compatible
from .tx_type import TaggedAddr
from decodex.utils import fmt_addr


@dataclass(frozen=True)
class ContractCreation(Action):
    deployer: TaggedAddr
    contract: TaggedAddr

    action: str = "contract_creation"

    def __str__(self):
        return f"{fmt_addr(self.deployer)} deployed contract {fmt_addr(self.contract)}"


@dataclass(frozen=True)
class UTF8Message(Action):
    """
    Represents a UTF8 message send from one address to another.
    """

    sender: TaggedAddr
    receiver: TaggedAddr
    message: str

    action: str = "utf8_message"

    def __str__(self):
        return f"{fmt_addr(self.sender)} sent message to {fmt_addr(self.receiver)}: {self.message}"


@dataclass(frozen=True)
class TransferAction(Action):
    """
    Represents a transfer of tokens from one address to another.
    """

    sender: TaggedAddr
    receiver: TaggedAddr
    token: ERC20Compatible
    amount: float

    action: str = "transfer"

    def __str__(self):
        return f"{fmt_addr(self.sender)} transferred {self.amount} {fmt_addr(self.token)} to {fmt_addr(self.receiver)}"


@dataclass(frozen=True)
class SwapAction(Action):
    """
    Represents a swap action between two tokens.
    """

    pool: TaggedAddr
    pay_token: ERC20Compatible
    recv_token: ERC20Compatible
    pay_amount: float
    recv_amount: float

    action: str = "swap"

    def __str__(self) -> str:
        tmpl = "Swap {pay_amount} {pay_token} for {recv_amount} {recv_token} on {pool}"
        return tmpl.format(
            pool=fmt_addr(self.pool),
            pay_amount=self.pay_amount,
            pay_token=fmt_addr(self.pay_token),
            recv_amount=self.recv_amount,
            recv_token=fmt_addr(self.recv_token),
        )


@dataclass(frozen=True)
class AddLiquidityAction(Action):
    """
    Represents an action to add liquidity to a pool.
    """

    pool: TaggedAddr
    token_0: ERC20Compatible
    token_1: ERC20Compatible
    amount_0: float
    amount_1: float

    action: str = "add_liquidity"

    def __str__(self) -> str:
        tmpl = "Add {amount_0} {token_0} and {amount_1} {token_1} liquidity to {pool}"
        return tmpl.format(
            amount_1=fmt_addr(self.amount_1),
            token_1=fmt_addr(self.token_1),
            pool=fmt_addr(self.pool),
            amount_0=self.amount_0,
            token_0=self.token_0,
        )


@dataclass(frozen=True)
class RemoveLiquidityAction(Action):
    """
    Represents an action to remove liquidity from a pool.
    """

    pool: TaggedAddr
    token_0: ERC20Compatible
    token_1: ERC20Compatible
    amount_0: float
    amount_1: float

    action: str = "remove_liquidity"

    def __str__(self) -> str:
        tmpl = "Remove {amount_0} {token_0} and {amount_1} {token_1} liquidity from {pool}"
        return tmpl.format(
            token_0=fmt_addr(self.token_0),
            token_1=fmt_addr(self.token_1),
            pool=fmt_addr(self.pool),
            amount_0=self.amount_0,
            amount_1=self.amount_1,
        )


@dataclass(frozen=True)
class PoolCreatedAction(Action):
    """
    Represents an action to create a pair.
    """

    token_0: ERC20Compatible
    token_1: ERC20Compatible
    fee: float = None

    action: str = "pool_created"

    def __str__(self) -> str:
        tmpl = "Create {token_0} / {token_1} pool"
        if not self.fee:
            return tmpl.format(
                token_0=fmt_addr(self.token_0),
                token_1=fmt_addr(self.token_1),
            )
        else:
            return tmpl + " with fee {fee}".format(
                token_0=fmt_addr(self.token_0),
                token_1=fmt_addr(self.token_1),
                fee=self.fee,
            )


@dataclass(frozen=True)
class CollectAction(Action):
    """
    Represents an action to collect tokens from a pool.
    """

    pool: TaggedAddr
    token_0: ERC20Compatible
    token_1: ERC20Compatible
    amount_0: float
    amount_1: float

    action: str = "collect"

    def __str__(self) -> str:
        tmpl = "Collect {amount_0} {token_0} and {amount_1} {token_1} from {pool}"
        return tmpl.format(
            token_0=fmt_addr(self.token_0),
            token_1=fmt_addr(self.token_1),
            pool=fmt_addr(self.pool),
            amount_0=self.amount_0,
            amount_1=self.amount_1,
        )


@dataclass(frozen=True)
class OwnerChangedAction(Action):
    """
    Represents an action to change the owner of a pool.
    """

    pool: TaggedAddr
    new_owner: TaggedAddr
    old_owner: TaggedAddr

    action: str = "owner_changed"

    def __str__(self) -> str:
        tmpl = "Change owner of {pool} from {old_owner} to {new_owner}"
        return tmpl.format(
            pool=fmt_addr(self.pool),
            old_owner=fmt_addr(self.old_owner),
            new_owner=fmt_addr(self.new_owner),
        )


@dataclass(frozen=True)
class BorrowAction(Action):
    """
    Represents an action to borrow tokens from a lending pool.
    """

    pool: TaggedAddr
    token: ERC20Compatible
    amount: float

    action: str = "borrow"

    def __str__(self) -> str:
        tmpl = "Borrow {amount} {token} from {pool}"
        return tmpl.format(
            amount=self.amount,
            token=fmt_addr(self.token),
            pool=fmt_addr(self.pool),
        )


@dataclass(frozen=True)
class RepayAction(Action):
    """
    Represents an action to repay tokens to a lending pool.
    """

    pool: TaggedAddr
    token: ERC20Compatible
    amount: float

    action: str = "repay"

    def __str__(self) -> str:
        tmpl = "Repay {amount} {token} to {pool}"
        return tmpl.format(
            amount=self.amount,
            token=fmt_addr(self.token),
            pool=fmt_addr(self.pool),
        )


@dataclass(frozen=True)
class DepositAction(Action):
    """
    Deposit tokens to a lending pool.
    """

    pool: TaggedAddr
    token: ERC20Compatible
    amount: float

    action: str = "deposit"

    def __str__(self) -> str:
        tmpl = "Deposit {amount} {token} to {pool}"
        return tmpl.format(
            amount=self.amount,
            token=fmt_addr(self.token),
            pool=fmt_addr(self.pool),
        )


@dataclass(frozen=True)
class WithdrawAction(Action):
    """
    Withdraw tokens from a lending pool.
    """

    pool: TaggedAddr
    token: ERC20Compatible
    amount: float
    receiver: TaggedAddr = None

    action: str = "withdraw"

    def __str__(self) -> str:
        if not self.receiver:
            tmpl = "Withdraw {amount} {token} from {pool}"
            return tmpl.format(
                amount=self.amount,
                token=fmt_addr(self.token),
                pool=fmt_addr(self.pool),
            )
        else:
            tmpl = "Withdraw {amount} {token} from {pool} to {receiver}"
            return tmpl.format(
                amount=self.amount,
                token=fmt_addr(self.token),
                pool=fmt_addr(self.pool),
            )


@dataclass(frozen=True)
class FlashloanAction(Action):
    """
    Represents an action to flashloan tokens from a lending pool.
    """

    pool: TaggedAddr
    token: ERC20Compatible
    amount: float

    action: str = "flashloan"

    def __str__(self) -> str:
        tmpl = "Flashloan {amount} {token} from {pool}"
        return tmpl.format(
            amount=self.amount,
            token=fmt_addr(self.token),
            pool=fmt_addr(self.pool),
        )


@dataclass(frozen=True)
class EnableCollateralAction(Action):
    """
    Enable a token as collateral in a lending pool.
    """

    pool: TaggedAddr
    token: ERC20Compatible

    action: str = "enable_collateral"

    def __str__(self) -> str:
        tmpl = "Enable {token} as collateral in {pool}"
        return tmpl.format(
            token=fmt_addr(self.token),
            pool=fmt_addr(self.pool),
        )


@dataclass(frozen=True)
class DisableCollateralAction(Action):
    """
    Disable a token as collateral in a lending pool.
    """

    pool: TaggedAddr
    token: ERC20Compatible

    action: str = "disable_collateral"

    def __str__(self) -> str:
        tmpl = "Disable {token} as collateral in {pool}"
        return tmpl.format(
            token=fmt_addr(self.token),
            pool=fmt_addr(self.pool),
        )


@dataclass(frozen=True)
class SupplyAction(Action):
    """
    Supply tokens to a lending pool.
    """

    pool: TaggedAddr
    token: ERC20Compatible
    amount: float

    action: str = "supply"

    def __str__(self) -> str:
        tmpl = "Supply {amount} {token} to {pool}"
        return tmpl.format(
            token=fmt_addr(self.token),
            pool=fmt_addr(self.pool),
            amount=self.amount,
        )



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional


@dataclass(frozen=True)
class Action:
    @property
    @abstractmethod
    def action(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    def dict(self):
        return vars(self)


EventHandleFunc = Callable[[Dict[str, Any]], Optional[Action]]



from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from multicall import Call
from multicall import Multicall

from decodex.convert.address import AddrTagger
from decodex.convert.signature import SignatureLookUp
from decodex.convert.token import ERC20TokenService
from decodex.type import Action
from decodex.type import AddLiquidityAction
from decodex.type import CollectAction
from decodex.type import EventHandleFunc
from decodex.type import EventPayload
from decodex.type import OwnerChangedAction
from decodex.type import PoolCreatedAction
from decodex.type import RemoveLiquidityAction
from decodex.type import SwapAction


class Events:
    def __init__(
        self,
        mc: Multicall,
        tagger: AddrTagger,
        *args,
        **kwargs,
    ) -> None:
        self._mc = mc
        self._tagger = tagger
        self._erc20_svc = ERC20TokenService(mc)


class UniswapEvents(Events):
    def _get_token_pair(self, pool_addr: str) -> Tuple[str, str]:
        result = self._mc.agg(
            [
                Call(
                    target=pool_addr,
                    function="token0()(address)",
                    request_id="token0",
                ),
                Call(
                    target=pool_addr,
                    function="token1()(address)",
                    request_id="token1",
                ),
            ]
        )
        token_map = {item["request_id"]: item["result"] for item in result}
        return token_map["token0"], token_map["token1"]


class UniswapV2Events(UniswapEvents):
    def __init__(
        self,
        mc: Multicall,
        tagger: AddrTagger,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(mc, tagger, *args, **kwargs)

    def swap(self) -> Tuple[str, EventHandleFunc]:
        # Swap(address indexed sender,uint amount0In, uint amount1In, uint amount0Out, uint amount1Out, address indexed to);
        text_sig = "Swap(address,uint256,uint256,uint256,uint256,address)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token0_addr, token1_addr = self._get_token_pair(payload["address"])
            token0, token1 = self._erc20_svc.batch_get_erc20([token0_addr, token1_addr])
            if token0 is None or token1 is None:
                return None
            params = payload["params"]
            amount0_diff = int(params["amount0Out"]) - int(params["amount0In"])
            amount1_diff = int(params["amount1Out"]) - int(params["amount1In"])
            if amount0_diff > 0:
                # pay token0, get token1
                pool, token0, token1 = self._tagger([payload["address"], token0, token1])
                return SwapAction(
                    pool=pool,
                    pay_token=token0,
                    pay_amount=abs(amount0_diff / 10 ** token0["decimals"]),
                    recv_token=token1,
                    recv_amount=abs(amount1_diff / 10 ** token1["decimals"]),
                )
            elif amount1_diff > 0:
                # pay token1, get token0
                pool, token0, token1 = self._tagger([payload["address"], token0, token1])
                return SwapAction(
                    pool=pool,
                    pay_token=token1,
                    pay_amount=abs(amount1_diff / 10 ** token1["decimals"]),
                    recv_token=token0,
                    recv_amount=abs(amount0_diff / 10 ** token0["decimals"]),
                )
            else:
                return None

        return text_sig, decoder

    def mint(self) -> Tuple[str, EventHandleFunc]:
        # Mint(address indexed sender, uint amount0, uint amount1);
        text_sig = "Mint(address,uint256,uint256)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token0_addr, token1_addr = self._get_token_pair(payload["address"])
            token0, token1 = self._erc20_svc.batch_get_erc20([token0_addr, token1_addr])
            if token0 is None or token1 is None:
                return None
            params = payload["params"]
            pool, token0, token1 = self._tagger([payload["address"], token0, token1])
            return AddLiquidityAction(
                pool=pool,
                token_0=token0,
                token_1=token1,
                amount_0=int(params["amount0"]) / 10 ** token0["decimals"],
                amount_1=int(params["amount1"]) / 10 ** token1["decimals"],
            )

        return text_sig, decoder

    def burn(self) -> Tuple[str, EventHandleFunc]:
        # Burn(address indexed sender, uint amount0, uint amount1, address indexed to);
        text_sig = "Burn(address,uint256,uint256,address)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            token0_addr, token1_addr = self._get_token_pair(payload["address"])
            params = payload["params"]
            token0, token1 = self._erc20_svc.batch_get_erc20([token0_addr, token1_addr])
            if token0 is None or token1 is None:
                return None
            pool, token0, token1 = self._tagger([payload["address"], token0, token1])
            return RemoveLiquidityAction(
                pool=pool,
                token_0=token0,
                token_1=token1,
                amount_0=int(params["amount0"]) / 10 ** token0["decimals"],
                amount_1=int(params["amount1"]) / 10 ** token1["decimals"],
            )

        return text_sig, decoder

    def pair_created(self) -> Tuple[str, EventHandleFunc]:
        # PairCreated(address indexed token0, address indexed token1, address pair, uint);
        text_sig = "PairCreated(address,address,address,uint256)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            params = payload["params"]
            token0, token1 = self._erc20_svc.batch_get_erc20([params["token0"], params["token1"]])
            if token0 is None or token1 is None:
                return None
            token0, token1 = self._tagger([token0, token1])
            return PoolCreatedAction(
                token_0=token0,
                token_1=token1,
            )

        return text_sig, decoder


class UniswapV3Events(UniswapEvents):
    def __init__(
        self,
        mc: Multicall,
        tagger: AddrTagger,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(mc, tagger, *args, **kwargs)

    def _get_tokens_by_position(self, pool_addr: str, pos_id: int) -> Tuple[str, str]:
        result = self._mc.agg(
            [
                Call(
                    target=pool_addr,
                    function="positions(uint256)(uint96,address,address,address,uint24,int24,int24,uint128,uint256,uint256,uint128,uint128)",
                    args=[pos_id],
                    request_id="positions",
                ),
            ]
        )
        if len(result) != 1:
            raise ValueError(f"Cannot find position {pos_id} in {pool_addr}")
        result = result[0]["result"]
        token0_addr, token1_addr = result[2], result[3]
        return token0_addr, token1_addr

    def pool_created(self) -> Tuple[str, EventHandleFunc]:
        # PoolCreated(address token0,address token1,uint24 fee,int24 tickSpacing,address pool)
        text_sig = "PoolCreated(address,address,uint24,int24,address)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            params = payload["params"]
            token0, token1 = self._erc20_svc.batch_get_erc20([params["token0"], params["token1"]])
            if token0 is None or token1 is None:
                return None
            token0, token1 = self._tagger([token0, token1])
            return PoolCreatedAction(
                token_0=token0,
                token_1=token1,
                fee=int(params["fee"]),
            )

        return text_sig, decoder

    def increase_liquidity(self) -> Tuple[str, EventHandleFunc]:
        # IncreaseLiquidity(uint256 indexed tokenId, uint128 liquidity, uint256 amount0, uint256 amount1);
        text_sig = "IncreaseLiquidity(uint256,uint128,uint256,uint256)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            params = payload["params"]
            token0_addr, token1_addr = self._get_tokens_by_position(payload["address"], int(params["tokenId"]))
            token0, token1 = self._erc20_svc.batch_get_erc20([token0_addr, token1_addr])
            if token0 is None or token1 is None:
                return None
            pool, token0, token1 = self._tagger([payload["address"], token0, token1])
            return AddLiquidityAction(
                pool=pool,
                token_0=token0,
                token_1=token1,
                amount_0=int(params["amount0"]) / 10 ** token0["decimals"],
                amount_1=int(params["amount1"]) / 10 ** token1["decimals"],
            )

        return text_sig, decoder

    def decrease_liquidity(self) -> Tuple[str, EventHandleFunc]:
        # DecreaseLiquidity(uint256 indexed tokenId, uint128 liquidity, uint256 amount0, uint256 amount1);
        text_sig = "DecreaseLiquidity(uint256,uint128,uint256,uint256)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            params = payload["params"]
            token0_addr, token1_addr = self._get_tokens_by_position(payload["address"], int(params["tokenId"]))
            token0, token1 = self._erc20_svc.batch_get_erc20([token0_addr, token1_addr])
            if token0 is None or token1 is None:
                return None
            pool, token0, token1 = self._tagger([payload["address"], token0, token1])
            return RemoveLiquidityAction(
                pool=pool,
                token_0=token0,
                token_1=token1,
                amount_0=abs(int(params["amount0"]) / 10 ** token0["decimals"]),
                amount_1=abs(int(params["amount1"]) / 10 ** token1["decimals"]),
            )

        return text_sig, decoder

    def swap(self) -> Tuple[str, EventHandleFunc]:
        # Swap(address sender,address recipient,int256 amount0,int256 amount1,uint160 sqrtPriceX96,uint128 liquidity,int24 tick)
        text_sig = "Swap(address,address,int256,int256,uint160,uint128,int24)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            params = payload["params"]
            token0_addr, token1_addr = self._get_token_pair(payload["address"])
            token0, token1 = self._erc20_svc.batch_get_erc20([token0_addr, token1_addr])
            if token0 is None or token1 is None:
                return None
            amount0, amount1 = int(params["amount0"]), int(params["amount1"])

            if amount0 > 0 and amount1 < 0:
                # pay token0, get token1
                pool, token0, token1 = self._tagger([payload["address"], token0, token1])
                return SwapAction(
                    pool=pool,
                    pay_token=token0,
                    pay_amount=abs(amount0 / 10 ** token0["decimals"]),
                    recv_token=token1,
                    recv_amount=abs(amount1 / 10 ** token1["decimals"]),
                )
            elif amount1 > 0 and amount0 < 0:
                # pay token1, get token0
                pool, token0, token1 = self._tagger([payload["address"], token0, token1])
                return SwapAction(
                    pool=pool,
                    pay_token=token1,
                    pay_amount=abs(amount1 / 10 ** token1["decimals"]),
                    recv_token=token0,
                    recv_amount=abs(amount0 / 10 ** token0["decimals"]),
                )
            else:
                return None

        return text_sig, decoder

    def collect(self) -> Tuple[str, EventHandleFunc]:
        # Collect(address owner,int24 tickLower,int24 tickUpper,uint128 amount0,uint128 amount1)
        text_sig = "Collect(address,int24,int24,uint128,uint128)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            params = payload["params"]
            token0_addr, token1_addr = self._get_token_pair(
                payload["address"],
            )
            token0, token1 = self._erc20_svc.batch_get_erc20([token0_addr, token1_addr])
            if token0 is None or token1 is None:
                return None
            amount0, amount1 = int(params["amount0"]), int(params["amount1"])
            pool, token0, token1 = self._tagger([payload["address"], token0, token1])
            return CollectAction(
                pool=pool,
                token_0=token0,
                token_1=token1,
                amount_0=abs(amount0 / 10 ** token0["decimals"]),
                amount_1=abs(amount1 / 10 ** token1["decimals"]),
            )

        return text_sig, decoder

    def owner_changed(self) -> Tuple[str, EventHandleFunc]:
        # OwnerChanged(address oldOwner, address newOwner)
        text_sig = "OwnerChanged(address,address)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            params = payload["params"]
            pool, new_owner, old_owner = self._tagger([payload["address"], params["newOwner"], params["oldOwner"]])
            return OwnerChangedAction(
                pool=pool,
                new_owner=new_owner,
                old_owner=old_owner,
            )

        return text_sig, decoder


class BancorEV3Events(Events):
    def __init__(
        self,
        mc: Multicall,
        tagger: AddrTagger,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(mc, tagger, *args, **kwargs)

    def tokens_traded(self) -> Tuple[str, EventHandleFunc]:
        # TokensTraded (index_topic_1 bytes32 contextId, index_topic_2 address sourceToken, index_topic_3 address targetToken, uint256 sourceAmount, uint256 targetAmount, uint256 bntAmount, uint256 targetFeeAmount, uint256 bntFeeAmount, address trader)
        text_sig = "TokensTraded(bytes32,address,address,uint256,uint256,uint256,uint256,uint256,address)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            params = payload["params"]
            token0_addr, token1_addr = params["sourceToken"], params["targetToken"]
            token0, token1 = self._erc20_svc.batch_get_erc20([token0_addr, token1_addr])
            if token0 is None or token1 is None:
                return None
            amount0, amount1 = int(params["sourceAmount"]), int(params["targetAmount"])
            pool, token0, token1 = self._tagger([payload["address"], token0, token1])
            if amount0 > 0 and amount1 < 0:
                # pay token0, get token1
                return SwapAction(
                    pool=pool,
                    pay_token=token0,
                    pay_amount=abs(amount0 / 10 ** token0["decimals"]),
                    recv_token=token1,
                    recv_amount=abs(amount1 / 10 ** token1["decimals"]),
                )
            elif amount1 > 0 and amount0 < 0:
                # pay token1, get token0
                return SwapAction(
                    pool=pool,
                    pay_token=token1,
                    pay_amount=abs(amount1 / 10 ** token1["decimals"]),
                    recv_token=token0,
                    recv_amount=abs(amount0 / 10 ** token0["decimals"]),
                )
            else:
                return None

        return text_sig, decoder


class CurveV2Events(Events):
    def __init__(
        self,
        mc: Multicall,
        tagger: AddrTagger,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(mc, tagger, *args, **kwargs)

    def token_exchange(self) -> Tuple[str, EventHandleFunc]:
        # TokenExchange (index_topic_1 address buyer, index_topic_2 address receiver, index_topic_3 address pool, address token_sold, address token_bought, uint256 amount_sold, uint256 amount_bought)
        text_sig = "TokenExchange(address,address,address,address,address,uint256,uint256)"

        def decoder(payload: EventPayload) -> Optional[Action]:
            params = payload["params"]
            token_sold, token_bought = self._erc20_svc.batch_get_erc20([params["token_sold"], params["token_bought"]])
            if token_sold is None or token_bought is None:
                return None
            pool, token_sold, token_bought = self._tagger([params["pool"], token_sold, token_bought])
            amout_sold, amout_bought = int(params["amount_sold"]), int(params["amount_bought"])
            return SwapAction(
                pool=pool,
                pay_token=token_sold,
                pay_amount=amout_sold / 10 ** token_sold["decimals"],
                recv_token=token_bought,
                recv_amount=abs(amout_bought) / 10 ** token_bought["decimals"],
            )

        return text_sig, decoder



from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import TypedDict

from .base import Action


Log = TypedDict(
    "Log",
    {
        "address": str,  # contract address, hex string. 0x prefixed.
        "topics": List[str],  # topics of the log, each is a hex string of 0x prefixed.
        "data": str,  # data of the log, hex string. 0x prefixed.
    },
)


TaggedAddr = TypedDict(
    "TaggedAddr",
    {
        "address": str,  # address or ens name
        "name": str,  # name of the address in our label dataset (e.g., USD Coin (USDC))
        "labels": List[str],  # label of the address (e.g., ["centre", "stablecoin"])
    },
)


class ERC20Compatible(TaggedAddr):
    contract_name: Optional[str]  # name of the contract, different from 'name' (e.g., USD Coin)
    decimals: Optional[int]  # decimals of the token (e.g., 6)
    symbol: Optional[str]  # symbol of the token (e.g., USDC)


AssetBalanceChanged = TypedDict(
    "BalanceChange",
    {
        "asset": ERC20Compatible,  # address of the asset
        "balance_change": float,  # balance change of the asset
    },
)


AccountBalanceChanged = TypedDict(
    "AccountBalanceChanged",
    {
        "address": TaggedAddr,  # address of the account
        "assets": List[AssetBalanceChanged],  # balance change of each asset
    },
)


Tx = TypedDict(
    "Tx",
    {
        "txhash": str,  # transaction hash, hex string. 0x prefixed.
        "from": str,  # from address, hex string. 0x prefixed.
        "to": Optional[str],  # to address, hex string. 0x prefixed.
        "contract_created": Optional[TaggedAddr],  # contract created by the transaction, hex string. 0x prefixed.
        "block_number": int,  # block number of the transaction
        "block_timestamp": int,  # timestamp of the block, in seconds.
        "value": int,  # value of the transaction, in wei.
        "gas_used": int,  # gas used by the transaction, in wei.
        "gas_price": int,  # gas price of the transaction, in wei.
        "input": str,  # input data of the transaction, hex string. 0x prefixed.
        "status": int,  # status of the transaction
        "reason": str,  # reason of the transaction if failed
        "logs": List[Log],  # logs of the transaction
        "eth_balance_changes": Dict[str, Dict[Literal["ETH", "Gas Fee"], int]],  # ETH balance change of the transaction
    },
    total=False,
)


TaggedTx = TypedDict(
    "TaggedTx",
    {
        "txhash": str,  # transaction hash, hex string. 0x prefixed.
        "from": TaggedAddr,  # from address, hex string. 0x prefixed.
        "to": Optional[TaggedAddr],  # to address, hex string. 0x prefixed.
        "contract_created": Optional[TaggedAddr],  # contract created by the transaction, hex string. 0x prefixed.
        "block_number": int,  # block number of the transaction
        "block_time": datetime,  # datetime of the block
        "value": float,  # value of the transaction, in Ether.
        "gas_used": int,  # gas used by the transaction, in Gwei.
        "gas_price": float,  # gas price of the transaction, in Gwei.
        "input": str,  # input data of the transaction, hex string. 0x prefixed.
        "status": int,  # status of the transaction
        "reason": str,  # reason of the transaction if failed
        "method": str,  # calling method of the transaction
        "actions": List[Action],  # actions of the transaction
        "balance_change": List[AccountBalanceChanged],  # balance change of the transaction
    },
    total=False,
)


EventPayload = TypedDict(
    "EventPayload",
    {
        "address": str,  # contract address, hex string. 0x prefixed.
        "params": Dict[str, Any],  # event parameters, the output from decoder
    },
)



from .action_type import AddLiquidityAction
from .action_type import BorrowAction
from .action_type import CollectAction
from .action_type import ContractCreation
from .action_type import DepositAction
from .action_type import DisableCollateralAction
from .action_type import EnableCollateralAction
from .action_type import FlashloanAction
from .action_type import OwnerChangedAction
from .action_type import PoolCreatedAction
from .action_type import RemoveLiquidityAction
from .action_type import RepayAction
from .action_type import SupplyAction
from .action_type import SwapAction
from .action_type import TransferAction
from .action_type import UTF8Message
from .action_type import WithdrawAction
from .base import Action
from .base import EventHandleFunc
from .rpc_type import RawTraceCallResponse
from .rpc_type import RawTraceCallResult
from .tx_type import AccountBalanceChanged
from .tx_type import AssetBalanceChanged
from .tx_type import ERC20Compatible
from .tx_type import EventPayload
from .tx_type import Log
from .tx_type import TaggedAddr
from .tx_type import TaggedTx
from .tx_type import Tx

__all__ = [
    "Tx",
    "Log",
    "TaggedTx",
    "ERC20Compatible",
    "TaggedAddr",
    "EventHandleFunc",
    "EventPayload",
    "Action",
    "ContractCreation",
    "TransferAction",
    "UTF8Message",
    "SwapAction",
    "AddLiquidityAction",
    "RemoveLiquidityAction",
    "PoolCreatedAction",
    "CollectAction",
    "OwnerChangedAction",
    "BorrowAction",
    "RepayAction",
    "DepositAction",
    "WithdrawAction",
    "FlashloanAction",
    "EnableCollateralAction",
    "DisableCollateralAction",
    "SupplyAction",
    "RpcRequest",
    "RawTraceCallResponse",
    "RawTraceCallResult",
    "AssetBalanceChanged",
    "AccountBalanceChanged",
]
